import sqlparse
from pyiceberg.expressions import *
from pyiceberg.expressions import parser

from duckberg.table import TableWithAlias


class DuckBergSQLParser:
    def parse_first_query(self, sql: str) -> sqlparse.sql.Statement:
        reformated_sql = sql.replace("'", '"')  # replace all single quotes with double quotes
        return sqlparse.parse(reformated_sql)[0]

    def unpack_identifiers(self, token: sqlparse.sql.IdentifierList) -> list[TableWithAlias]:
        return list(
            map(
                lambda y: TableWithAlias.from_identifier(y),
                filter(lambda x: type(x) is sqlparse.sql.Identifier, token.tokens),
            )
        )

    def extract_tables(self, parsed_sql: sqlparse.sql.Statement) -> list[TableWithAlias]:
        tables = []
        get_next = 0
        c_table: list[TableWithAlias] = []
        c_table_wc = None
        for token in parsed_sql.tokens:
            if get_next == 1 and token.ttype is not sqlparse.tokens.Whitespace:
                if type(token) is sqlparse.sql.Identifier:
                    c_table = [TableWithAlias.from_identifier(token)]
                    get_next += 1
                elif type(token) is sqlparse.sql.IdentifierList:
                    c_table = self.unpack_identifiers(token)
                    get_next += 1
                elif type(token) is sqlparse.sql.Parenthesis:
                    tables.extend(self.extract_tables(token))

            if token.ttype is sqlparse.tokens.Keyword and str(token.value).upper() == "FROM":
                get_next += 1

            if type(token) is sqlparse.sql.Where:
                c_table_wc = self.extract_where_conditions(token)

        mapped_c_table = list(map(lambda x: x.set_comparisons(c_table_wc), c_table))
        tables.extend(mapped_c_table)
        return tables

    def extract_where_conditions(self, where_statement: list[sqlparse.sql.Where]):
        comparison = sqlparse.sql.TokenList(where_statement[1:])
        return parser.parse(str(comparison))



import os

from duckberg import DuckBerg

os.environ["AWS_DEFAULT_REGION"] = "eu-west-1"
os.environ["AWS_ACCESS_KEY_ID"] = "ASIARCC6UKQSEPHP3JGZ"
os.environ["AWS_SECRET_ACCESS_KEY"] = "D7oKLEDjuduNvh9kx4vlkThb+lNfnibtWISVwLkR"
os.environ[
    "AWS_SESSION_TOKEN"
] = "FwoGZXIvYXdzEPD//////////wEaDDM8aIRHZ+nHFdccHiKJAi/gWd/fGJrUkEgi97YgF0Iiy6may1PlGxucXKpps6/61rofuO2fuq0sb49tdHsYiJAohu7Hua8crG8Vqz93EY/2H9Is48OsSeQqi7cj3z0drRsWHReqJVJdAg4q6TV4U20l5XH4Fh+9yNx8ip3z0dCXNPdrO8KPRKEwlgwd9s0kUFMGPtGuKg3aUHwZWFeiLsbSFkpgyFOF4rSWIw2uLMTSXIh7Yhu68B28fZDc2gL34fgefR/aBdGan1Czn76K0YRRu400TXMFbL+MnELCiDb3uf5vEzL+blAd6AYZLARU+OnRKzZrwjxgp7F65Z9rYfdiIHT5UGoU5AwNx1tXcmG8j2517piXfAoolr/LrQYyKw/UfgzuMrwk/5DlTAZvDMWoTLApnLbndbSkzkJyrbAQlClmgV9raA9YOy0="


MINIO_URI = "http://localhost:9000/"
MINIO_USER = "admin"
MINIO_PASSWORD = "password"

# catalog_config: dict[str, str] = {
#     "type": "rest",
#     "uri": "http://localhost:8181/",
#     "credentials": "admin:password",
#     "s3.endpoint": MINIO_URI,
#     "s3.access-key-id": MINIO_USER,
#     "s3.secret-access-key": MINIO_PASSWORD,
# }

# BOTO_SESSION_CONFIG_KEYS = ["aws_access_key_id", "aws_secret_access_key", "aws_session_token", "region_name", "profile_name"]

catalog_config: dict[str, str] = {
    "type": "glue",
}

catalog_name = "default"

db = DuckBerg(
    catalog_name=catalog_name, catalog_config=catalog_config, database_names=["slido_iceberg"], table_names=["datapi_questions_iceberg"]
)

tables = db.list_tables()

# New way of quering data without partition filter
query: str = "SELECT * FROM 'slido_iceberg.datapi_questions_iceberg'"
df = db.select(
    sql=query,
    table="slido_iceberg.datapi_questions_iceberg",
    partition_filter="organization_id = '4ad6a753-3252-49bd-9f5d-adaa150ad58a'",
).read_pandas()

print(len(df))

# assert(df['count_star()'][0] == 2614)
#
# # New way of quering data
# query: str = "SELECT count(*) FROM (SELECT * FROM 'nyc.taxis' WHERE payment_type = 1 AND trip_distance > 40 ORDER BY tolls_amount DESC)"
# df = db.select(sql=query).read_pandas()
# assert(df['count_star()'][0] == 1673)
#
# # Old way of quering data
# query: str = "SELECT count(*) FROM (SELECT * FROM 'nyc.taxis' WHERE payment_type = 1 AND trip_distance > 40 ORDER BY tolls_amount DESC)"
# df = db.select(sql=query, table="nyc.taxis", partition_filter="payment_type = 1").read_pandas()
# assert(df['count_star()'][0] == 1673)



from duckberg import DuckBerg

MINIO_URI = "http://localhost:9000/"
MINIO_USER = "admin"
MINIO_PASSWORD = "password"

# catalog_config: dict[str, str] = {
#     "type": "rest",
#     "uri": "http://localhost:8181/",
#     "credentials": "admin:password",
#     "s3.endpoint": MINIO_URI,
#     "s3.access-key-id": MINIO_USER,
#     "s3.secret-access-key": MINIO_PASSWORD,
# }

BOTO_SESSION_CONFIG_KEYS = [
    "aws_access_key_id",
    "aws_secret_access_key",
    "aws_session_token",
    "region_name",
    "profile_name",
]

catalog_config: dict[str, str] = {
    "type": "glue",
    "region_name": "eu-west-1",
    "AWS_ACCESS_KEY_ID": "ASIARCC6UKQSJHYFTYU3",
    "AWS_SECRET_ACCESS_KEY": "GOVQOPw/+4uDIiOTyBBaCd5XIiwSi4/L8d81kWKE",
    "AWS_SESSION_TOKEN": "FwoGZXIvYXdzEIj//////////wEaDGBJbs1mx0RRY+WdiyKJAhPVsk7CNbXKBrJwhPW4djBKdon/+3HACgtvRoQmkRNp2A+TiF7zisUk1Fk2NLP23i3AKA1KsPmBS1LylIAHiFAhyij/xegAHRdELLpY8ftvH9q35/7I3eAT7TNjf2q0RSpW1y9/1BT52p87Bmmkj/FYpWQrnTZcGAESHRJJGHwmXwqzTBs93FtdBn+FvRs3ivi+zl3KWe+vc/mXUfgfV3NixjSdu7QuvcuzmL6xylEyyhlUpCu4eikyQ6EFEgYrUdGhBNaiyTtJ9v8tSmurvYh5K04L9aOQXoKnHxZ3KOXUDKMga7vhbnY4bH1dAqocVYD7rV45HLMumbpfwwqw+S8hjJtFL+ThDoUoqPGLrAYyK7LfqaHrcMSg39YFegYdkir6FcDma+Vv1XfjrAY/qwFVDhWqGpfWucRprgQ=",
}

catalog_name = "default"

db = DuckBerg(catalog_name=catalog_name, catalog_config=catalog_config, database_names=["slido_iceberg"])

tables = db.list_tables()

assert len(tables) == 1

# New way of quering data without partition filter
query: str = (
    "SELECT count(*) FROM (SELECT * FROM 'slido_iceberg.taxis' WHERE trip_distance > 40 ORDER BY tolls_amount DESC)"
)
df = db.select(sql=query).read_pandas()
assert df["count_star()"][0] == 2614

# New way of quering data
query: str = "SELECT count(*) FROM (SELECT * FROM 'nyc.taxis' WHERE payment_type = 1 AND trip_distance > 40 ORDER BY tolls_amount DESC)"
df = db.select(sql=query).read_pandas()
assert df["count_star()"][0] == 1673

# Old way of quering data
query: str = "SELECT count(*) FROM (SELECT * FROM 'nyc.taxis' WHERE payment_type = 1 AND trip_distance > 40 ORDER BY tolls_amount DESC)"
df = db.select(sql=query, table="nyc.taxis", partition_filter="payment_type = 1").read_pandas()
assert df["count_star()"][0] == 1673



from duckberg.sqlparser import DuckBergSQLParser

parser = DuckBergSQLParser()


sql1 = """
SELECT * FROM this_is_awesome_table WHERE a > 15"""
sql1_parsed = parser.parse_first_query(sql=sql1)
res1 = parser.extract_tables(sql1_parsed)
res1_where = str(res1[0].comparisons)
assert "GreaterThan(term=Reference(name='a'), literal=LongLiteral(15))" == res1_where

sql2 = """
SELECT * FROM this_is_awesome_table WHERE a > 15 AND a < 16"""
sql2_parsed = parser.parse_first_query(sql=sql2)
res2 = parser.extract_tables(sql2_parsed)
res2_where = str(res2[0].comparisons)
assert (
    "And(left=GreaterThan(term=Reference(name='a'), literal=LongLiteral(15)), right=LessThan(term=Reference(name='a'), literal=LongLiteral(16)))"
    == res2_where
)

sql3 = """
SELECT * FROM this_is_awesome_table WHERE (a > 15 AND a < 16) OR c > 15"""
sql3_parsed = parser.parse_first_query(sql=sql3)
res3 = parser.extract_tables(sql3_parsed)
res3_where = str(res3[0].comparisons)
assert (
    "Or(left=And(left=GreaterThan(term=Reference(name='a'), literal=LongLiteral(15)), right=LessThan(term=Reference(name='a'), literal=LongLiteral(16))), right=GreaterThan(term=Reference(name='c'), literal=LongLiteral(15)))"
    == res3_where
)



from duckberg.sqlparser import DuckBergSQLParser

parser = DuckBergSQLParser()


sql1 = """
SELECT * FROM this_is_awesome_table"""
sql1_parsed = parser.parse_first_query(sql=sql1)
res1 = parser.extract_tables(sql1_parsed)
assert len(res1) == 1
assert list(map(lambda x: str(x), res1)) == ["this_is_awesome_table (None)"]

sql2 = """
SELECT * FROM this_is_awesome_table, second_awesome_table"""
sql2_parsed = parser.parse_first_query(sql=sql2)
res2 = parser.extract_tables(sql2_parsed)
assert len(res2) == 2
assert list(map(lambda x: str(x), res2)) == ["this_is_awesome_table (None)", "second_awesome_table (None)"]

sql3 = """
SELECT * FROM (SELECT * FROM (SELECT * FROM this_is_awesome_table))"""
sql3_parsed = parser.parse_first_query(sql=sql3)
res3 = parser.extract_tables(sql3_parsed)
assert len(res3) == 1
assert list(map(lambda x: str(x), res3)) == ["this_is_awesome_table (None)"]

sql4 = """
SELECT * FROM (SELECT * FROM (SELECT * FROM this_is_awesome_table), second_awesome_table)"""
sql4_parsed = parser.parse_first_query(sql=sql4)
res4 = parser.extract_tables(sql4_parsed)
assert len(res4) == 2
assert list(map(lambda x: str(x), res4)) == ["this_is_awesome_table (None)", "second_awesome_table (None)"]

sql5 = """
SELECT * FROM (SELECT * FROM (SELECT * FROM this_is_awesome_table tiat, second_awesome_table))"""
sql5_parsed = parser.parse_first_query(sql=sql5)
res5 = parser.extract_tables(sql5_parsed)
assert len(res5) == 2
assert list(map(lambda x: str(x), res5)) == ["this_is_awesome_table (tiat)", "second_awesome_table (None)"]



"""Module containing services needed for executing queries with Duckdb + Iceberg."""

import re
from typing import Optional

import duckdb
from pyarrow.lib import RecordBatchReader
from pyiceberg.catalog import Catalog, load_catalog
from pyiceberg.expressions import AlwaysTrue

from duckberg.exceptions import TableNotInCatalogException
from duckberg.sqlparser import DuckBergSQLParser
from duckberg.table import DuckBergTable, TableWithAlias

BATCH_SIZE_ROWS = 1024
DEFAULT_MEM_LIMIT = "1GB"
DEFAULT_DB_THREAD_LIMIT = 1
DEFAULT_CATALOG_NAME = "default"


class DuckBerg:
    """The query engine for easy querying Iceberg data through DuckDb.

    Args:
        catalog_name: The name of the catalog.
        catalog_config: The catalog config that is used for PyIceberg init.
        database_names: The list of database names that are used to filter just databases we need.
        table_names: The list of table names that are planned to be used. This list is useful to filter out the tables
            that are not Iceberg tables.
    """

    def __init__(
        self,
        catalog_name: str,
        catalog_config: dict[str, str],
        database_names: str,
        table_names: Optional[str] = None,
        duckdb_connection: duckdb.DuckDBPyConnection = None,
        db_thread_limit: Optional[int] = DEFAULT_DB_THREAD_LIMIT,
        db_mem_limit: Optional[str] = DEFAULT_MEM_LIMIT,
        batch_size_rows: Optional[int] = BATCH_SIZE_ROWS,
    ):
        self.db_thread_limit = db_thread_limit
        self.db_mem_limit = db_mem_limit
        self.batch_size_rows = batch_size_rows
        self.duckdb_connection = duckdb_connection

        if self.duckdb_connection is None:
            self.duckdb_connection = duckdb.connect()
            self.__init_duckdb()

        self.sql_parser = DuckBergSQLParser()
        self.tables: dict[str, DuckBergTable] = {}

        self.__init_duckdb()
        self.__init_tables(catalog_config, catalog_name, database_names, table_names)

    def __init_duckdb(self):
        self.duckdb_connection.execute(f"SET memory_limit='{self.db_mem_limit}'")
        self.duckdb_connection.execute(f"SET threads TO {self.db_thread_limit}")

    def __init_tables(self, catalog_config, catalog_name, database_names, table_names):
        self.tables = {}

        catalog: Catalog = load_catalog(catalog_name, **catalog_config)

        namespaces = catalog.list_namespaces()
        namespaces = [i for i in namespaces if i[0] in database_names]

        tables = {}
        for n in namespaces:
            tables_list = catalog.list_tables(n)
            if table_names:
                tables_list = [t for t in tables_list if t[1] in table_names]
            tables: dict[str, DuckBergTable] = {
                ".".join(t): DuckBergTable.from_pyiceberg_table(catalog.load_table(t)) for t in tables_list
            }
            self.tables = self.tables | tables

    def list_tables(self):
        return list(self.tables.keys())

    def list_partitions(self, table: str):
        t = self.tables[table]

        if t.partitions is None:
            t.precomp_partitions()

        return t.partitions

    def select(
        self, sql: str, table: str = None, partition_filter: str = None, sql_params: [str] = None
    ) -> RecordBatchReader:
        if table is not None and partition_filter is not None:
            return self._select_old(sql, table, partition_filter, sql_params)

        parsed_sql = self.sql_parser.parse_first_query(sql)
        extracted_tables = self.sql_parser.extract_tables(parsed_sql)

        table: TableWithAlias
        for table in extracted_tables:
            table_name = table.table_name

            if table_name not in self.tables:
                raise TableNotInCatalogException

            row_filter = AlwaysTrue()
            if table.comparisons is not None:
                row_filter = table.comparisons

            table_data_scan_as_arrow = self.tables[table_name].scan(row_filter=row_filter).to_arrow()
            self.duckdb_connection.register(table_name, table_data_scan_as_arrow)

        if sql_params is None:
            return self.duckdb_connection.execute(sql).fetch_record_batch(self.batch_size_rows)
        else:
            return self.duckdb_connection.execute(sql, parameters=sql_params).fetch_record_batch(self.batch_size_rows)

    def _select_old(self, sql: str, table: str, partition_filter: str, sql_params: [str] = None):
        table_data_scan_as_arrow = self.tables[table].scan(row_filter=partition_filter).to_arrow()
        self.duckdb_connection.register(table, table_data_scan_as_arrow)

        if sql_params is None:
            return self.duckdb_connection.execute(sql).fetch_record_batch(self.batch_size_rows)
        else:
            return self.duckdb_connection.execute(sql, parameters=sql_params).fetch_record_batch(self.batch_size_rows)



import sqlparse
from pyiceberg.catalog import Catalog
from pyiceberg.expressions import BooleanExpression
from pyiceberg.io import FileIO
from pyiceberg.table import Table
from pyiceberg.table.metadata import TableMetadata
from pyiceberg.typedef import Identifier


class DuckBergTable(Table):
    """
    Class for storing precomputed data for faster processing of queries
    """

    def __init__(
        self, identifier: Identifier, metadata: TableMetadata, metadata_location: str, io: FileIO, catalog: Catalog
    ) -> None:
        super().__init__(identifier, metadata, metadata_location, io, catalog)
        self.partitions = None

    @classmethod
    def from_pyiceberg_table(cls, table: Table):
        return cls(table.identifier, table.metadata, table.metadata_location, table.io, table.catalog)

    def precomp_partitions(self):
        if self.spec().is_unpartitioned():
            self.partitions = []

        partition_cols_ids = [p["source-id"] for p in self.spec().model_dump()["fields"]]
        self.partitions = [c["name"] for c in self.schema().model_dump()["fields"] if c["id"] in partition_cols_ids]

    def __repr__(self) -> str:
        return self.table


class TableWithAlias:
    """
    Dataclass contains table name with alias
    """

    def __init__(self, tname: str, talias: str) -> None:
        self.table_name: str = tname
        self.table_alias: str = talias
        self.comparisons: BooleanExpression = None

    @classmethod
    def from_identifier(cls, identf: sqlparse.sql.Identifier):
        return cls(identf.get_real_name(), identf.get_alias())

    def set_comparisons(self, comparisons: BooleanExpression):
        self.comparisons = comparisons
        return self

    def __str__(self) -> str:
        return f"{self.table_name} ({self.table_alias})"

    def __repr__(self) -> str:
        return self.__str__()
